import math
import re
from typing import Union, List, Tuple, Dict

import gym

from pandas_ml_common import pd, np
from pandas_ml_quant_rl.cache import NoCache
from pandas_ml_quant_rl.cache.abstract_cache import Cache
from pandas_ml_utils import FeaturesAndLabels
from pandas_ml_quant_rl.model.environment import Environment, Strategy
from .utils import  SpaceUtils


class RandomAssetEnv(Environment):

    def __init__(self,
                 features_and_labels: FeaturesAndLabels,
                 symbols: Union[str, List[str]],
                 strategy: Strategy,
                 pct_train_data: float = 0.8,
                 max_steps: int = None,
                 min_training_samples: float = np.inf,
                 use_cache: Cache = NoCache(),
                 ):
        super().__init__(strategy)
        self.max_steps = math.inf if max_steps is None else max_steps
        self.min_training_samples = min_training_samples
        self.features_and_labels = features_and_labels
        self.pct_train_data = pct_train_data

        # define spaces
        self.observation_space = None

        # define execution mode
        self.cache = use_cache
        self.mode = 'train'
        self.done = True

        # load symbols available to randomly draw from
        if isinstance(symbols, str):
            with open(symbols) as f:
                self.symbols = np.array(re.split("[\r\n]|[\n]|[\r]", f.read()))
        else:
            self.symbols = symbols

        # finally make a dummy initialisation
        self._init()

    def as_train(self) -> Tuple['RandomAssetEnv', Tuple[pd.DataFrame, np.ndarray]]:
        # note that shallow copy is shuffling due to the _init call
        copy = self._shallow_copy()
        copy.mode = 'train'
        return copy, self._current_state()

    def as_test(self) -> Tuple['RandomAssetEnv', Tuple[pd.DataFrame, np.ndarray]]:
        # note that shallow copy is shuffling due to the _init call
        copy = self._shallow_copy()
        copy.mode = 'test'
        return copy, self._current_state()

    def as_predict(self) -> Tuple['RandomAssetEnv', Tuple[pd.DataFrame, np.ndarray]]:
        # note that shallow copy is shuffling due to the _init call
        copy = self._shallow_copy()
        copy.mode = 'predict'
        return copy, self._current_state()

    def _shallow_copy(self):
        # note that shallow copy is shuffling due to the _init call
        return RandomAssetEnv(
            self.features_and_labels,
            self.symbols,
            self.strategy,
            self.pct_train_data,
            self.max_steps,
            self.min_training_samples,
            self.cache
        )

    def sample_action(self, probs=None):
        return self.strategy.sample_action(probs)

    def step(self, action) -> Tuple[Tuple[np.ndarray, np.ndarray], float, bool, Dict]:
        old_price_frame = self._current_price_frame()

        # regardless what action we take we will end up in the next time step
        # and this is the first possible time where we actually can do something
        self._state_idx += 1
        new_price_frame = self._current_price_frame()

        # let the agent execute an action according to our strategy
        strategy_state, reward, game_over = self.strategy.trade_reward(
            old_price_frame,
            action,
            new_price_frame
        )

        # check if the epoch is over
        self.done = game_over \
                    or self._state_idx >= self._last_index \
                    or self._state_idx > (self._start_idx + self.max_steps)

        # return new state reward and done flag
        return self._current_state(strategy_state), reward, self.done, {}

    def render(self, mode='human'):
        return self.strategy.render(mode)

    def reset(self) -> Tuple[np.ndarray, np.ndarray]:
        self.strategy.reset()
        return self._init()

    def _init(self) -> Tuple[np.ndarray, np.ndarray]:
        # this is reinforcement learning, we can only extract features (no labels)!
        self._symbol = np.random.choice(self.symbols, 1).item()
        self._df = self.cache.get_data_or_fetch(self._symbol)
        self._features, self._features_index = self.cache.get_feature_frames_or_fetch(self._df, self._symbol, self.features_and_labels)
        self._is_feature_tuple = isinstance(self._features, tuple)
        nr_of_samples = len(self._features[0]) if self._is_feature_tuple else len(self._features)

        if self.mode in ['train', 'test']:
            if self.mode == 'train':
                self._last_index = int(nr_of_samples * 0.8)
                # allow at least one step
                self._state_idx = np.random.randint(1, self._last_index - 1, 1).item()
            if self.mode == 'test':
                self._last_index = nr_of_samples
                # if min samples is infinity then we start from the first index of the test data
                test_start_idx = int(nr_of_samples * 0.8)
                test_end_idx = min(nr_of_samples - self.min_training_samples, test_start_idx + 1)
                self._state_idx = np.random.randint(test_start_idx, test_end_idx, 1).item()
        else:
            self._last_index = nr_of_samples
            self._labels = pd.DataFrame({})
            self._sample_weights = pd.DataFrame({})
            self._gross_loss = pd.DataFrame({})
            self._state_idx = 0

        if self.observation_space is None:
            self.observation_space = gym.spaces.Tuple((
                SpaceUtils.unbounded_tuple_boxes(*[f.shape[1:] for f in self._features]) if self._is_feature_tuple else SpaceUtils.unbounded_box(self._features.shape[1:]),
                SpaceUtils.unbounded_box(self.strategy.current_state().shape)
            ))

        self._start_idx = self._state_idx
        self.done = self._state_idx >= self._last_index
        return self._current_state()

    def _current_state(self, strategy_state=None) -> Tuple[Union[np.ndarray, Tuple[np.ndarray]], np.ndarray]:
        # make sure to return it as a batch of size one therefore use list of state_idx
        return (
            tuple([f[[self._state_idx]] for f in self._features]) if self._is_feature_tuple else self._features[[self._state_idx]],
            strategy_state if strategy_state is not None else self.strategy.current_state()
        )

    def _current_price_frame(self) -> pd.Series:
        ts = self._features_index[self._state_idx]
        return self._df.loc[ts]


# ...
# TODO later allow features to be a list of feature sets echa witha possible different shape

# Also later we want to implement an envoronment where the agent has access to all assets and he needs to pick a combination
# of assets like in a portfolio construction manner



from typing import Tuple, Dict

import gym
import numpy as np

from pandas_ml_quant_rl.model.environment import Environment


class CartPoleWrappedEnv(Environment):

    def __init__(self, auto_render_after_steps=None):
        self._cart_pole_env = gym.make("CartPole-v0")
        self.action_space = self._cart_pole_env.action_space
        self.observation_space = self._cart_pole_env.observation_space
        self.auto_render_after_steps = auto_render_after_steps
        self._total_reward = 0
        self._step = 0

    def as_train(self) -> Tuple['Environment', Tuple[np.ndarray, np.ndarray]]:
        return self

    def as_test(self, renderer=None) -> Tuple['Environment', Tuple[np.ndarray, np.ndarray]]:
        return self

    def as_predict(self, renderer=None) -> Tuple['Environment', Tuple[np.ndarray, np.ndarray]]:
        return self

    def sample_action(self, probs=None):
        return np.random.choice(self._cart_pole_env.action_space.n, p=probs)

    def step(self, action) -> Tuple[Tuple[np.ndarray, np.ndarray], float, bool, Dict]:
        state, reward, done, info = self._cart_pole_env.step(action)
        self._total_reward += reward
        self._step += 1

        return (state, np.array([])), reward, done, info

    def reset(self) -> Tuple[np.ndarray, np.ndarray]:
        self._total_reward = 0
        self._step = 0

        state = self._cart_pole_env.reset()
        return (state, np.array([]))

    def get_screen(self):
        return self._cart_pole_env.render('rgb_array')


import os
from pandas_ml_common import pd


def load_symbol(symbol):
    return pd.read_csv(
        os.path.join(os.path.dirname(os.path.abspath(__file__)), ".data", f"{symbol}.csv"),
        index_col='Date',
        parse_dates=True
    )




from time import sleep
from typing import Dict, Tuple, List
import numpy as np
import matplotlib
import matplotlib.pyplot as plt
import matplotlib.dates as mdates

from pandas_ml_common.plot.utils import matplot_dates
from .abstract_renderer import Renderer


class CandleStickRenderer(Renderer):

    def __init__(self, action_mapping: List[Tuple[int, str, str]] = None, figsize=(20, 10)):
        super(CandleStickRenderer, self).__init__()
        matplotlib.use('Qt5Agg')
        plt.ion()

        self.action_map = action_mapping
        self.figsize = figsize

        self.min_time_step = 1.0
        self.r = 0

        self.reset()

    def reset(self):
        self.r = 0

        self.fig, self.axes = plt.subplots(2, 1, sharex=True, gridspec_kw={"height_ratios": [3, 1]}, figsize=self.figsize)
        self.fig.canvas.draw()  # draw and show it
        plt.show(block=False)

        for ax in self.axes:
            ax.xaxis_date()
            ax.xaxis.set_major_formatter(mdates.DateFormatter('%d-%m-%Y'))

    def plot(self, old_state, action, new_state, reward, done):
        df = new_state.to_frame().T
        x = matplot_dates(df)
        o = df["Open"].values
        h = df["High"].values
        l = df["Low"].values
        c = df["Close"].values

        # indicator x is half a time step before / after the trading day
        bx = x - (self.min_time_step * 0.4)
        sx = x + (self.min_time_step * 0.4)

        # plot candle
        b = min(o, c)
        color = 'red' if o > c else 'green'
        self.axes[0].vlines(x, l, h, color=color)
        self.axes[0].bar(x, max(o, c) - b, bottom=b, color=color)

        # plot actions
        for _action, plot_act, column in self.action_map:
            if _action == action:
                if plot_act == 'buy':
                    self.axes[0].scatter(bx, df[column], marker=">", color='green')
                if plot_act == 'sell':
                    self.axes[0].scatter(sx, df[column], marker="<", color='red')

        # plot reward
        self.r += reward.item() if isinstance(reward, np.ndarray) else float(reward)
        print(x, self.r)
        self.axes[1].bar(x, self.r, color='silver')

        if done:
            sleep(1)
            plt.close(self.fig)
            self.reset()

    def render(self, mode=None, min_time_step=1.0):
        self.min_time_step = min_time_step

        for ax in self.axes:
            ax.autoscale_view(tight=True, scalex=True, scaley=True)

        self.fig.canvas.draw()
        plt.pause(0.05)






from typing import Tuple

import numpy as np
from gym.spaces import Discrete

from pandas_ml_quant_rl.model.environment import Strategy


class BuyOpenSellCloseSellOpenBuyClose(Strategy):
    """
    This is kind of a dummy strategy set where we have 3 actions:
    0: do nothing
    1: buy at the open and sell the same day close
    2: sell the close and buy the same day open
    """

    def __init__(self,
                 long: Tuple[str, str] = ('Open', 'Close'),
                 short: Tuple[str, str] = ('Close', 'Open'),
                 max_loos: float = -0.2
                 ):
        super().__init__(Discrete(3), buffer=1, assets=1, indicators=1)
        self.long_columns = long
        self.short_columns = short
        self.max_loos = max_loos
        self.total_return = 0

    def sample_action(self, probs=None):
        action = np.random.choice(self.action_space.n, p=probs) if probs is not None else self.action_space.sample()
        return action

    def trade_reward(self, previous_bar, action, bar):
        if action == 0:
            reward = -0.00001
        elif action == 1:
            reward = np.log(bar[self.long_columns[1]] / bar[self.long_columns[0]])
        else:
            reward = np.log(bar[self.long_columns[0]] / bar[self.long_columns[1]])

        self.total_return += reward

        # remember one indicator for one asset
        row_to_buffer = np.array([[reward]])
        return self._roll_state_buffer(row_to_buffer), reward, self.total_return < self.max_loos

    def reset(self):
        super().reset()
        self.total_return = 0


class LongOnly(Strategy):

    def __init__(self,
                 buy: str = 'Close',
                 sell: str = 'Close',
                 max_loos: float = -0.2,
                 max_holding_period: int = None
                 ):
        # we have 4 actions hold, buy, sell, swing
        super().__init__(Discrete(4), buffer=1, assets=1, indicators=1)
        if max_holding_period is not None: raise NotImplemented()  # FIXME implement max holding period and eventuall also implement something like stop loss orders
        self.buy_col = buy
        self.sell_col = sell
        self.max_loos = max_loos
        self.max_holding_period = max_holding_period

        self.positioning = False
        self.total_return = 0
        self.opening_value = None

    def sample_action(self, probs=None):
        if probs is None: probs = np.ones(self.action_space.n)
        if not isinstance(probs, np.ndarray): probs = np.array(probs)

        if self.positioning:
            # we are long so buy is not allowed
            probs[1] = 0

        if not self.positioning:
            # we are not positioned so sell is not allowed
            probs[2] = 0

        # rescale probabilities
        probs = probs / probs.sum()

        action = np.random.choice(self.action_space.n, p=probs)
        return action

    def trade_reward(self, previous_bar, action, bar) -> Tuple[np.ndarray, float, bool]:
        # FIXME ....
        reward = 0
        row_to_buffer = np.array([[reward]])
        return self._roll_state_buffer(row_to_buffer), reward, self.total_return < self.max_loos

    def reset(self):
        super().reset()
        self.positioning = False
        self.total_return = 0
        self.opening_value = None


class LongShortSwing(Strategy):

    def __init__(self,
                 buy: str = 'Close',
                 sell: str = 'Close',
                 max_loos: float = -0.2,
                 max_holding_period: int = None
                 ):
        # we have 4 actions hold, buy, sell, swing
        super().__init__(Discrete(4), buffer=1, assets=1, indicators=1)
        if max_holding_period is not None: raise NotImplemented()  # FIXME implement max holding period and eventuall also implement something like stop loss orders
        self.buy_col = buy
        self.sell_col = sell
        self.max_loos = max_loos
        self.max_holding_period = max_holding_period

        self.positioning = 0
        self.total_return = 0
        self.opening_value = None

    def sample_action(self, probs=None):
        if probs is None: probs = np.ones(self.action_space.n)
        if not isinstance(probs, np.ndarray): probs = np.array(probs)

        if self.positioning > 0:
            # we are long so buy is not allowed
            probs[1] = 0
        elif self.positioning < 0:
            # we are short so sell is not allowed
            probs[2] = 0
        elif self.positioning == 0:
            # we cant swing if we don't have an open position yet
            probs[3] = 0

        # rescale probabilities
        probs = probs / probs.sum()

        action = np.random.choice(self.action_space.n, p=probs)
        return action

    def trade_reward(self, previous_bar, action, bar):
        if action == 1 and self.positioning != 1:
            # enter a new long position, reward is 0
            self.opening_value = bar[self.buy_col]
            if self.positioning < 0:
                self.positioning = 0
                reward = np.log(self.opening_value / bar[self.buy_col])
            else:
                self.positioning = 1
                reward = 0
        elif action == 2 and self.positioning != -1:
            # enter a new long position, reward is 0
            self.opening_value = bar[self.sell_col]
            if self.positioning > 0:
                self.positioning = 0
                reward = np.log(bar[self.sell_col] / self.opening_value)
            else:
                self.positioning = -1
                reward = 0
        elif action == 3 and self.positioning != 0:
            # reward is revenue of closing the position
            if self.positioning > 0:
                pct = (bar[self.sell_col] / self.opening_value)
                self.opening_value = bar[self.sell_col]
            else:
                pct = (self.opening_value / bar[self.buy_col])
                self.opening_value = bar[self.buy_col]

            self.positioning = -self.positioning
            reward = np.log(pct)
        else:
            # reward is the position value change or -0.00001 in case of no open position
            # if we exceed the maximum holding period we should also use the closing posiiton reward
            if self.positioning == 0:
                reward = -0.0001
            else:
                pct = (bar[self.sell_col] / self.opening_value) if self.positioning > 0 else (self.opening_value / bar[self.buy_col])
                reward = np.log(pct)

        self.total_return += reward

        # remember our current positioning -> we should use this information in the network such that the agent
        # learns which actions are executable given a long/sort position
        return self._roll_state_buffer(np.array([[self.positioning]])), reward, self.total_return < self.max_loos

    def reset(self):
        super().reset()
        self.positioning = 0
        self.total_return = 0
        self.opening_value = None


# TODO eventually support stop orders for risk mamagement
#  reward could also be sharp ratio or sortino ratio
#  use the log return change of the portfolio net worth, eventually only use excess log returns (demean the return)



import time
from queue import Empty
from multiprocessing import Process, Event, Queue
from .abstract_renderer import Renderer


def render_frame(data_q: Queue, finish_e: Event, renderer_provider):
    renderer = renderer_provider()

    while True:
        try:
            old_state, action, new_state, reward, done = data_q.get(timeout=0.1)
            renderer.plot(old_state, action, new_state, reward, done)
            renderer.render()
        except Empty:
            renderer.render()
            if finish_e.wait(0.1):
                break

    print("shut down online rendering !!!")


class OnlineRenderer(Renderer):

    def __init__(self, renderer_provider):
        super().__init__()
        self.data_q = Queue()
        self.finish_e = Event()
        self.worker = Process(target=render_frame, args=(self.data_q, self.finish_e, renderer_provider))
        self.startup = True

    def plot(self, old_state, action, new_state, reward, done):
        self.data_q.put_nowait((old_state, action, new_state, reward, done))
        if self.startup:
            time.sleep(1)
            self.startup = False

    def stop(self):
        self.finish_e.set()

    def render(self, mode=None, min_time_step=1.0):
        if not self.worker.is_alive():
            self.worker.start()


class MovieRenderer(Renderer):

    def __init__(self, renderer_provider):
        """
        import numpy as np
        import matplotlib
        matplotlib.use("Agg")
        import matplotlib.pyplot as plt
        import matplotlib.animation as manimation

        FFMpegWriter = manimation.writers['ffmpeg']
        metadata = dict(title='Movie Test', artist='Matplotlib',
                        comment='Movie support!')
        writer = FFMpegWriter(fps=15, metadata=metadata)

        fig = plt.figure()
        l, = plt.plot([], [], 'k-o')

        plt.xlim(-5, 5)
        plt.ylim(-5, 5)

        x0, y0 = 0, 0

        with writer.saving(fig, "writer_test.mp4", 100):
            for i in range(100):
                x0 += 0.1 * np.random.randn()
                y0 += 0.1 * np.random.randn()
                l.set_data(x0, y0)
                writer.grab_frame()
        :param renderer_provider:
        """
        pass





class Renderer(object):

    def __init__(self):
        pass

    def plot(self, old_state, action, new_state, reward, done):
        pass

    def render(self, mode=None, min_time_step=1.0):
        pass






from unittest import TestCase

from pandas_ml_quant import pd, np
from pandas_ml_quant_rl.cache import NoCache, FileCache
from pandas_ml_quant_rl.cache.mem_cache import MemCache
from pandas_ml_quant_rl_test.config import load_symbol
from pandas_ml_utils import FeaturesAndLabels


class TestFileCache(TestCase):

    def test_file_cache(self):
        fl = FeaturesAndLabels(
            features=[lambda df: df["Close"].ta.sma()]
        )

        cache1 = NoCache(load_symbol)
        cache2 = FileCache('/tmp/cache.test', load_symbol)

        for i in range(2):
            pd.testing.assert_frame_equal(
                cache1.get_data_or_fetch("SPY"),
                cache2.get_data_or_fetch("SPY")
            )

            np.testing.assert_almost_equal(
                cache1.get_feature_frames_or_fetch(cache1.get_data_or_fetch("SPY"), "SPY", fl)[0],
                cache2.get_feature_frames_or_fetch(cache1.get_data_or_fetch("SPY"), "SPY", fl)[0]
            )

    def test_file_cache_with_multi_frame(self):
        fl = FeaturesAndLabels(
            features=(
                [lambda df: df["Close"].ta.sma()],
                [lambda df: df["Close"].ta.macd()]
            )
        )

        cache1 = NoCache(load_symbol)
        cache2 = FileCache('/tmp/cache.test', load_symbol)

        for i in range(2):
            pd.testing.assert_frame_equal(
                cache1.get_data_or_fetch("SPY"),
                cache2.get_data_or_fetch("SPY")
            )

            np.testing.assert_almost_equal(
                cache1.get_feature_frames_or_fetch(cache1.get_data_or_fetch("SPY"), "SPY", fl)[0][0],
                cache2.get_feature_frames_or_fetch(cache1.get_data_or_fetch("SPY"), "SPY", fl)[0][0]
            )

    def test_mem_cache(self):
        fl = FeaturesAndLabels(
            features=[lambda df: df["Close"].ta.sma()]
        )

        cache1 = NoCache(load_symbol)
        cache2 = MemCache(load_symbol)

        for i in range(2):
            pd.testing.assert_frame_equal(
                cache1.get_data_or_fetch("SPY"),
                cache2.get_data_or_fetch("SPY")
            )

            np.testing.assert_almost_equal(
                cache1.get_feature_frames_or_fetch(cache1.get_data_or_fetch("SPY"), "SPY", fl)[0],
                cache2.get_feature_frames_or_fetch(cache1.get_data_or_fetch("SPY"), "SPY", fl)[0]
            )

    def test_mem_cache_with_multi_frame(self):
        fl = FeaturesAndLabels(
            features=(
                [lambda df: df["Close"].ta.sma()],
                [lambda df: df["Close"].ta.macd()]
            )
        )

        cache1 = NoCache(load_symbol)
        cache2 = MemCache(load_symbol)

        for i in range(2):
            pd.testing.assert_frame_equal(
                cache1.get_data_or_fetch("SPY"),
                cache2.get_data_or_fetch("SPY")
            )

            np.testing.assert_almost_equal(
                cache1.get_feature_frames_or_fetch(cache1.get_data_or_fetch("SPY"), "SPY", fl)[0][0],
                cache2.get_feature_frames_or_fetch(cache1.get_data_or_fetch("SPY"), "SPY", fl)[0][0]
            )



from time import sleep

from pandas_ml_quant import PostProcessedFeaturesAndLabels
from pandas_ml_quant_rl.cache import FileCache
from pandas_ml_quant_rl.environments import RandomAssetEnv
from pandas_ml_quant_rl.model.strategies.discrete import BuyOpenSellCloseSellOpenBuyClose
from pandas_ml_quant_rl.renderer import CandleStickRenderer, OnlineRenderer
from pandas_ml_quant_rl_test.config import load_symbol


def render_from_environment():

    env = RandomAssetEnv(
        PostProcessedFeaturesAndLabels(
            features=[
                lambda df: df["Close"].ta.log_returns()
            ],
            labels=[
                lambda df: df["Close"].ta.log_returns().shift(-1)
            ],
            feature_post_processor=[
                lambda df: df.ta.rnn(60)
            ],
        ),
        ["SPY", "GLD"],
        pct_train_data=0.8,
        max_steps=10,
        use_cache=FileCache('/tmp/lalala.hd5', load_symbol),
        strategy=BuyOpenSellCloseSellOpenBuyClose(),
        renderer=OnlineRenderer(
            lambda: CandleStickRenderer
                (action_mapping=[(1, 'buy', 'Open'), (1, 'sell', 'Close'), (2, 'sell', 'Open'), (2, 'buy', 'Close')])
        )
    )

    env.render()
    for _ in range(2):
        print("play episode")
        env.reset()
        done = False
        while not done:
            s, r, done, _ = env.step(env.action_space.sample())
            print(s.index, done)

    sleep(2)
    env.renderer.stop()


if __name__ == '__main__':
    render_from_environment()



import numpy as np


class Policy(object):

    def __init__(self):
        self.is_learning_mode = True

    def eval(self):
        self.is_learning_mode = False
        return self

    def train(self):
        self.is_learning_mode = True
        return self

    def reset(self):
        pass

    def choose_action(self, env, state, render_on_axis=None):
        pass

    def log_experience(self, state, action, reward, new_state, done, info):
        if self.is_learning_mode:
            self.learn(state, action, reward, new_state, done, info)

    def learn(self, state, action, reward, new_state, done, info):
        pass





from typing import Callable, Tuple

from pandas_ml_common import Typing, pd, np
from pandas_ml_utils.ml.data.extraction import extract_features
from .abstract_cache import Cache


class MemCache(Cache):

    def __init__(self, data_provider: Callable[[str], Typing.PatchedDataFrame] = None):
        super().__init__(data_provider)
        self.data = {}
        self.features = {}

    def get_data_or_fetch(self, symbol) -> Typing.PatchedDataFrame:
        if symbol not in self.data:
            self.data[symbol] = self.data_provider(symbol)

        return self.data[symbol]

    def get_feature_frames_or_fetch(self, df, symbol, features_and_labels) -> Tuple[np.ndarray, pd.Index]:
        if symbol not in self.features:
            _, features, _ = extract_features(df, features_and_labels)
            self.features[symbol] = (features._.values, features.index)

        return self.features[symbol]



from typing import Tuple, List
import numpy as np


class Buffer(object):

    def sample(self, size) -> Tuple[List, ...]:
        raise NotImplemented()

    def append_args(self, *row) -> 'Buffer':
        return self.append_row(row)

    def append_row(self, row: np.ndarray) -> 'Buffer':
        raise NotImplemented()

    def reset(self) -> 'Buffer':
        raise NotImplemented()

    def __getitem__(self, item):
        # return the column requested
        raise NotImplemented()

    def __len__(self):
        raise NotImplemented()


import gym
import numpy as np

class SpaceUtils(object):

    @staticmethod
    def unbounded_box(shape):
        return gym.spaces.Box(low=-np.inf, high=np.inf, shape=shape)

    @staticmethod
    def unbounded_tuple_boxes(*shapes):
        ts = tuple([gym.spaces.Box(low=-np.inf, high=np.inf, shape=shape) for shape in shapes])
        return gym.spaces.Tuple(ts)



import os
from typing import Callable

from pandas_ml_common import pd, Typing
from pandas_ml_common.decorator import MultiFrameDecorator
from pandas_ml_utils.ml.data.extraction import extract_features
from .abstract_cache import Cache


class FileCache(Cache):

    def __init__(self, file_name: str, data_provider: Callable[[str], Typing.PatchedDataFrame] = None):
        super().__init__(data_provider)
        # remove old cache file
        os.path.exists(file_name) and os.remove(file_name)
        self.file_cache = pd.HDFStore(file_name, mode='a')

    def get_data_or_fetch(self, symbol):
        if symbol in self.file_cache:
            return self.file_cache[symbol]
        else:
            df = self.data_provider(symbol)
            self.file_cache[symbol] = df
            return df

    def get_feature_frames_or_fetch(self, df, symbol, features_and_labels):
        fkey = f'{symbol}__features'

        # note! features can be a MultiFrameDecorator !!
        if isinstance(features_and_labels.features, tuple):
            fkeys = [f'{symbol}__features__{i}' for i in range(len(features_and_labels.features))]
            features = [self.file_cache[fkey] for fk in self.file_cache if fk in fkeys]

            if len(features) == len(fkeys):
                features = MultiFrameDecorator(features)
            else:
                _, features, _ = extract_features(df, features_and_labels)
                for fk, f in zip(fkeys, features.frames()):
                    self.file_cache[fk] = f
        else:
            if fkey in self.file_cache:
                features = self.file_cache[fkey]
            else:
                print(f"fetch data for: {symbol}")
                _, features, _ = extract_features(df, features_and_labels)
                self.file_cache[fkey] = features

        return features._.values, features.index




import torch as T


def grow_same_ndim(*tensor: T.tensor):
    max_dim = max([t.ndim for t in tensor])

    def grow(t):
        res = t
        while res.ndim < max_dim:
            res = res.unsqueeze(-1)

        return res

    return tuple([grow(t) for t in tensor])


from typing import Callable, Iterable

from .environment import Environment
from .policy import Policy


class Agent(object):

    def __init__(self, ema_factor=20, render_nr_actions=0, render_env=True, render_policy=True, figsize=(24, 8)):
        self.render_nr_actions = render_nr_actions
        self.ema_factor = 1 / (ema_factor + 1)
        self.render_env = render_env
        self.render_policy = render_policy

        nr_of_subplots = render_env + render_policy
        if nr_of_subplots > 0:
            import matplotlib.pyplot as plt

            if isinstance(nr_of_subplots, tuple):
                fig, ax = plt.subplots(*nr_of_subplots, figsize=figsize)
            else:
                fig, ax = plt.subplots(1, nr_of_subplots, figsize=figsize)

            self.fig = fig
            self.ax = ax if isinstance(ax, Iterable) else [ax]
        else:
            self.fig = None
            self.ax = (None, None)

    def learn_to_play(self, env: Environment, policy: Policy, exit_criteria: Callable[[float, int], bool]):
        policy.reset()
        return self._play_util(env.as_train(), policy.train(), exit_criteria)

    def play_one_episode(self, env: Environment, policy: Policy):
        episode_reward = self._play_episode(env.as_predict(), policy.eval())
        if self.fig is not None:
            import matplotlib.pyplot as plt
            self.fig.suptitle(f"reward {episode_reward}")
            self.fig.canvas.draw()
            plt.close(self.fig)

        return episode_reward

    def _play_util(self, env: Environment, policy: Policy, exit_criteria: Callable[[float, int], bool]):
        total_reward_ma = None
        total_reward = 0
        episodes = 0

        while True:
            episode_reward = self._play_episode(env, policy)
            total_reward += episode_reward
            episodes += 1

            if total_reward_ma is not None:
                total_reward_ma = episode_reward * self.ema_factor + (total_reward_ma * (1 - self.ema_factor))
            else:
                total_reward_ma = episode_reward

            if self.fig is not None:
                self.fig.suptitle(f"reward {episode_reward} / mean {total_reward_ma:.2f}")
                self.fig.canvas.draw()

            if exit_criteria(total_reward_ma, episodes):
                print(f"Solved {episode_reward} (mean {total_reward_ma}) in {episodes} episodes")
                break

        if self.fig is not None:
            import matplotlib.pyplot as plt
            plt.close(self.fig)

        return total_reward_ma

    def _play_episode(self, env: Environment, policy: Policy):
        # initialize our episode
        state = env.reset()
        episode_reward = 0
        action_counter = 0
        done = False

        while not done:
            # we render the environment and the policy
            do_render = self.render_nr_actions > 0 and action_counter % self.render_nr_actions == 0

            # we use the policy to provide an action
            action = policy.choose_action(env, state, self.ax[-1] if do_render else None)
            action_counter += 1

            # we execute the action in the environment
            new_state, reward, done, info = env.execute_action(action, self.ax[0] if do_render else None)
            episode_reward += reward

            # we trigger a learning function on the policy
            policy.log_experience(state, action, reward, new_state, done, info)
            state = new_state

            # if render draw canvas
            if do_render: self.fig.canvas.draw()

        # render last frame and return reward
        if self.fig is not None:
            env.render(self.ax[0])
            policy.choose_action(env, state, self.ax[-1])

        return episode_reward

# TODO add tensorboard logging
# todo add log chart for action space entropy
#         self.tensorboard_writer = EasySummaryWriter(self.__class__.__name__, **kwargs)
#    def log_to_tensorboard(self, cumulative_reward, nr_of_steps=None, action_hist=None, scalars=None):
#        # log rewards
#        if self._cumulative_reward_ma is None:
#            self._cumulative_reward_ma = cumulative_reward
#        else:
#            self._cumulative_reward_ma = cumulative_reward * (2 / 21) + (self._cumulative_reward_ma * (1 - 2 / 21))
#
#        self.tensorboard_writer.add_scalars(
#            "reward",
#            {"reward": cumulative_reward, "ma": self._cumulative_reward_ma},
#            self._episodes
#        )
#
#        # log nuber of steps per episode
#        if nr_of_steps is not None:
#            self.tensorboard_writer.add_scalar("nr of steps", nr_of_steps, self._episodes)
#
#        # log actions
#        if action_hist is not None:
#            action, counts = np.unique(action_hist, return_counts=True)
#            self.tensorboard_writer.add_scalars(
#                "action counts",
#                {str(a): c for a, c in zip(action, counts)},
#                self._episodes
#            )
#
#        # log epsilon
#        if scalars is not None:
#            for item, number in scalars.items():
#                self.tensorboard_writer.add_scalar(item, number, self._episodes)
#
#        # flush data and increase steps
#        self.tensorboard_writer.flush()
#        self._episodes += 1
#


from tensorboardX import SummaryWriter
import time

from pandas_ml_quant_rl.wrapper.tensorboardX import EasySummaryWriter

log_file_context = 'lala'
log_file = f'/tmp/{log_file_context}/{int(time.time() / 60) - int(time.time() / 60 / 100000) * 100000}'
print(f"tensorboard --logdir {log_file_context}")

#sw = SummaryWriter(log_file)
sw = EasySummaryWriter("test")

for i in range(100):
    sw.add_scalars("lala", {"a": i, "b": -i}, i)

#sw.add_histogram()


from functools import lru_cache
from typing import Tuple, Union, List

import numpy as np
import torch as T
import torch.nn as nn


class PolicyNetwork(nn.Module):

    def __init__(self):
        super().__init__()

    def forward(self, x: Union[List, Tuple[np.ndarray]], action=None, render_axis=None):
        if isinstance(x, tuple):
            # single sample
            return self.forward([x], [action] if action is not None else None, render_axis)
        else:
            # batch of samples
            state = self._unzip_to_tensor(x)
            if render_axis is not None:
                with T.no_grad():
                    self.render(state, render_axis)

            return self.get_state_value(state) if action is None else self.get_value_for(state, action)

    def _unzip_to_tensor(self, rows):
        # unzip a batch of samples
        if len(rows) < 0:
            # in case of an empty batch return an empty tensor
            return T.FloatTensor([]).to(self.device)

        columns = []
        nr_of_rows = len(rows)
        nr_of_columns = len(rows[0])

        for j in range(nr_of_columns):
            if isinstance(rows[0][j], tuple):
                # if in the first row a column is a tuple we need to unnest these tuples: [([...], [...]), ...]
                list_of_tuples = [row[j] for row in rows]
                columns.append(self._unzip_to_tensor(list_of_tuples))
            else:
                # the values in this column is not a tuple: [[...], ...] so we can stack all of then
                list_of_values = [row[j] for row in rows]
                columns.append(T.FloatTensor(np.vstack(list_of_values)).to(self.device))

        return tuple(columns)

    @property
    @lru_cache(1)
    def device(self):
        devices = ({param.device for param in self.parameters()} |
                   {buf.device for buf in self.buffers()})

        if len(devices) != 1:
            raise RuntimeError('Cannot determine device: {} different devices found'
                               .format(len(devices)))

        return next(iter(devices))

    def get_value_for(self, states: Tuple[List[np.ndarray], ...], action: List) -> T.tensor:
        raise NotImplementedError()

    def get_state_value(self, states: Tuple[List[np.ndarray], ...]) -> Tuple[T.tensor, T.tensor]:
        raise NotImplementedError()

    def render(self, state, ax):
        pass



import multiprocessing as mp
import random
import time
from queue import Empty

import matplotlib.pyplot as plt
import numpy


def worker(q):
    #plt.ion()
    fig=plt.figure()
    ax=fig.add_subplot(111)
    ln, = ax.plot([], [])
    fig.canvas.draw()   # draw and show it
    plt.show(block=False)
    x = 0

    while True:
        try:
            obj = q.get(timeout = 0.8)
            n = obj + 0
            x += 1
            print("sub : got:", n)

            ax.bar([x], [n])

        except Empty as e:
            ax.autoscale_view(tight=True, scalex=True, scaley=False)
            fig.canvas.draw()
            plt.pause(0.05)
            continue


if __name__ == '__main__':
    queue = mp.Queue()
    p = mp.Process(target=worker, args=(queue,))
    p.start()

    while True:
        n = random.random() * 5
        queue.put(n)
        time.sleep(1.0)




from typing import Callable, Tuple

import yfinance

from pandas_ml_common import pd, Typing


class Cache(object):

    def __init__(self, data_provider: Callable[[str], Typing.PatchedDataFrame] = None):
        self.data_provider = yfinance.download if data_provider is None else data_provider

    def get_data_or_fetch(self, symbol) -> pd.DataFrame:
        pass

    def get_feature_frames_or_fetch(self, df, symbol, features_and_labels) -> Tuple[pd.DataFrame, ...]:
        pass



from typing import Callable, Tuple

from pandas_ml_common import Typing
from pandas_ml_utils.ml.data.extraction import extract_features
from .abstract_cache import Cache
import numpy as np
import pandas as pd


class NoCache(Cache):

    def __init__(self, data_provider: Callable[[str], Typing.PatchedDataFrame] = None):
        super().__init__(data_provider)

    def get_data_or_fetch(self, symbol) -> Typing.PatchedDataFrame:
        return self.data_provider(symbol)

    def get_feature_frames_or_fetch(self, df, symbol, features_and_labels) -> Tuple[np.ndarray, pd.Index]:
        _, features, self._targets = extract_features(df, features_and_labels)
        features_index = features.index
        features = features._.values

        return features, features_index



from unittest import TestCase
import numpy as np


from pandas_ml_quant_rl.buffer.list_buffer import ListBuffer


class TestBuffers(object):

    def init_buffer(self, buffer, ignore_capacity=False):
        for i in range(9):
            buffer.append_row(np.array([i, 10 - i]))

        buffer.append_row(np.array([9, 0]))

        return buffer


class TestListBuffer(TestBuffers, TestCase):

    def test_init(self):
        self.init_buffer(ListBuffer(2), True)





import torch as T
import torch.nn as nn
import torch.functional as F


class QuantileHuberLoss(nn.Module):
    def __init__(self):
        super().__init__()
        self.quantile_tau = None

    def huber_loss(self, x, k=1.0):
        return T.where(x.abs() < k, 0.5 * x.pow(2), k * (x.abs() - 0.5 * k))

    def forward(self, input, target):
        assert input.ndim == 3, "expected shape (batch, quantile, action)"
        if self.quantile_tau is None:
            batch, quantile, action = input.shape
            self.quantile_tau = T.FloatTensor([i / quantile for i in range(1, quantile + 1)]).to(input.device)

        td_error = target - input
        huber_loss = self.huber_loss(td_error)
        quantile = abs(self.quantile_tau - (td_error.detach() < 0).float()) * huber_loss / 1.0
        loss = quantile.sum(dim=1).mean(dim=1)
        loss = loss.mean()
        return loss





from tensorboardX import SummaryWriter
from time import time
import shutil
import os


class EasySummaryWriter(SummaryWriter):

    def __init__(self, context, path='/tmp/tensorboard/', purge=True, **kwargs):
        print(f"tensorboard --logdir {path}")
        tensec = int(time() / 6)
        time_path = str(tensec - int(tensec / 100000.) * 100000)

        if purge:
            print(f"purge {path}")
            shutil.rmtree(path, ignore_errors=True)

        log_file = os.path.join(path, time_path, context)
        super().__init__(log_file, **kwargs)






from typing import Dict
from typing import Tuple

import gym
import numpy as np
import pandas as pd
from gym import Space
from matplotlib.figure import Figure

from pandas_ml_quant_rl.renderer.abstract_renderer import Renderer


class Strategy(object):

    def __init__(self,
                 action_space: Space,
                 buffer: int = 1,
                 assets: int = 1,
                 indicators: int = 1,
                 renderer: Renderer = Renderer()
                 ):
        """

        :param action_space: a gmy space defining the possible actions
        :param buffer: size of a buffer which is holding information like executed actions and rewards but
            also information like currently open positions for example. Each strategy needs to decide which information
            is needed. default buffer size = 1
        :param assets: number of assets simultaneously trade-able, default 1
        :param indicators: the number of indicators the buffer can hold like action and rewards would be 2.
            default = 1 (reward)
        """
        self.action_space = action_space
        self.buffer = buffer
        self.assets = assets
        self.indicators = indicators
        self.buffered_state = np.zeros((buffer, assets, indicators))

    def sample_action(self, *args, **kwargs):
        raise NotImplementedError()

    def current_state(self) -> np.ndarray:
        return self.buffered_state

    def trade_reward(self, previous_bar: pd.Series, action, bar: pd.Series) -> Tuple[np.ndarray, float, bool]:
        """

        :param action:
        :param label:
        :param sample_weight:
        :param gross_loss:
        :return: we return a tuple of (state, reward, bust)
        the sate of the portfolio after the action which is a 3D matrix of [timestepes, assets, indicators].
        Each strategy has to implement a set of indicators like rewards, positions weights or sharp ratio etc.

        during training we execute the action and calculate the immediate return of this action alongside
        the information if we can continue trading or not (i.e. we got bust).
        """

        return self.current_state(), 0, False

    def _roll_state_buffer(self, state: np.ndarray) -> np.ndarray:
        """
        A helper function to roll over the state array

        :return: rolled state
        """

        # rollover the current state and store the new data
        self.buffered_state = np.roll(self.buffered_state, -1, 0)
        self.buffered_state[-1] = state
        return self.buffered_state

    def current_available_actions(self) -> Tuple[int]:
        """

        :return: the set of actions available in the current state i.e. in a long only portfolio with a position only
        sell or hold are actual possible actions from this state
        """
        return self.action_space.shape

    def reset(self) -> np.ndarray:
        """
        When we start a new episode we need to reset the strategy parameters as well
        :return:
        """
        self.buffered_state = np.zeros(self.buffered_state.shape)
        return self.current_state()

    def render(self, mode='human'):
        return None


class Environment(gym.Env):

    def __init__(self, strategy: Strategy):
        super().__init__()
        self.strategy = strategy
        self.action_space = strategy.action_space

    def sample_action(self, probs=None):
        return self.strategy.sample_action(probs)

    def execute_action(self, action, render_on_axis=None) -> Tuple[Tuple[np.ndarray, np.ndarray], float, bool, Dict]:
        if render_on_axis is not None: self.render(render_on_axis)
        return self.step(action)

    def render(self, mode) -> Figure:
        import matplotlib

        if mode == 'matplotlib':
            import matplotlib.pyplot as plt
            return plt.imshow(self.get_screen())
        elif isinstance(mode, matplotlib.axes.Axes):
            ax = mode
            ax.clear()
            ax.imshow(self.get_screen())
            return ax

    def step(self, action) -> Tuple[Tuple[np.ndarray, np.ndarray], float, bool, Dict]:
        raise NotImplementedError()

    def as_train(self) -> Tuple['Environment', Tuple[np.ndarray, np.ndarray]]:
        raise NotImplementedError()

    def as_test(self, renderer=None) -> Tuple['Environment', Tuple[np.ndarray, np.ndarray]]:
        raise NotImplementedError()

    def as_predict(self, renderer=None) -> Tuple['Environment', Tuple[np.ndarray, np.ndarray]]:
        raise NotImplementedError()

    def reset(self) -> Tuple[np.ndarray, np.ndarray]:
        raise NotImplementedError()

    def get_screen(self):
        raise NotImplementedError()




from typing import Callable

import numpy as np
import torch as T
import torch.nn as nn
import torch.optim as optim

from pandas_ml_quant_rl.model.policy.pytorch.torch_network import PolicyNetwork
from pandas_ml_quant_rl.model.policy import Policy
from pandas_ml_quant_rl.model.buffer import ListBuffer
from pandas_ml_quant_rl.model.policy.pytorch.torch_utils import grow_same_ndim


class DQN_Policy(Policy):

    def __init__(self,
                 network: Callable[[], PolicyNetwork],
                 replay_buffer_size: int = 10000,
                 batch_size: int = 32,
                 gamma: float = 0.99,
                 epsilon: float = 1.0,
                 epsilon_decay: float = 100,
                 min_epsilon: float = 0.05,
                 nr_episodes_update_target: int = 10,
                 optimizer: Callable[[T.tensor], optim.Optimizer] = lambda params: optim.Adam(params=params, lr=0.01),
                 objective=nn.MSELoss(),
                 verbose=0,
                 **kwargs
                 ):
        super().__init__(**kwargs)
        self.network = network()
        self.target_network = network()
        self.device = self.network.device
        self.replay_buffer_size = replay_buffer_size
        self.batch_size = batch_size
        self.nr_episodes_update_target = nr_episodes_update_target
        self.gamma = gamma
        self.epsilon = epsilon
        self.epsilon_decay = epsilon_decay
        self.min_epsilon = min_epsilon
        self.optimizer = optimizer(self.network.parameters())
        self.objective = objective
        self.verbose = verbose

    def train(self):
        self.network.train()
        self.target_network.eval()
        return super().train()

    def eval(self):
        self.network.eval()
        self.target_network.eval()
        return super().eval()

    def reset(self):
        self._copy_weights()
        self.replay_buffer = ListBuffer(["state", "action", "reward", "state prime", "done"],
                                        max_size=self.replay_buffer_size)
        self.copy_target_cnt = 0
        self.nr_of_episodes = 0
        self.nr_of_batches = 0

    def choose_action(self, env, state, render_on_axis=None):
        with T.no_grad():
            action = self.network(state, render_axis=render_on_axis)[1].cpu().detach().item()

        if self.is_learning_mode:
            # decay epsilon after each batch of episodes
            epsilon = self.min_epsilon + (self.epsilon - self.min_epsilon) * np.exp(-1. * self.nr_of_batches / self.epsilon_decay)
            if np.random.random() < epsilon:
                return env.sample_action()

        return action

    def learn(self, state, action, reward, new_state, done, info):
        self.replay_buffer.append_args(state, action, reward, new_state, done)

        if len(self.replay_buffer) >= self.batch_size:
            # after we have reached a buffer size of batch size this is equivalent to the number of episodes
            self.nr_of_batches += 1

            # sample experiences from the replay buffer
            states, actions, rewards, next_states, dones = self.replay_buffer.sample(self.batch_size)

            # get the current Q values from the network
            current_q_values = self.network(states, actions)

            # get the target Q values for the "label"
            target_q_values = self._calc_target_q_values(next_states, rewards, dones, self.device)

            # back propagate and clamp overshooting gradients
            loss = self._backprop(current_q_values, target_q_values)
        if done:
            self.nr_of_episodes += 1

            # after n episodes we update our target network or perform a soft update
            if self.nr_episodes_update_target < 0:
                self._soft_update_weights()
            elif self.nr_of_episodes % self.nr_episodes_update_target == 0:
                self._copy_weights()
                self.copy_target_cnt += 1

    def _calc_target_q_values(self, next_states, rewards, dones, device):
        next_state_values = self.target_network(next_states)[0].detach()
        rewards = T.FloatTensor(rewards).to(device)
        dones = T.FloatTensor(dones).to(device)

        next_state_values, rewards, dones = grow_same_ndim(next_state_values, rewards, dones)
        target_q_values = rewards + (self.gamma * next_state_values * (1 - dones))
        return target_q_values

    def _backprop(self, current_q_values, target_q_values):
        input, target = grow_same_ndim(current_q_values, target_q_values)
        loss = self.objective(input, target)
        self.optimizer.zero_grad()
        loss.backward()

        for param in self.network.parameters():
            param.grad.data.clamp_(-1, 1)

        self.optimizer.step()
        return loss.detach().cpu().item()

    def _copy_weights(self):
        self.target_network.load_state_dict(self.network.state_dict())
        self.target_network.eval()

    def _soft_update_weights(self):
        local_model = self.network
        target_model = self.target_network.eval()
        TAU = self.nr_episodes_update_target

        for target_param, local_param in zip(target_model.parameters(), local_model.parameters()):
            target_param.data.copy_(TAU * local_param.data + (1.0 - TAU)*target_param.data)




from typing import Tuple, List, Union, Iterable
import numpy as np


class Buffer(object):

    def sample(self, size) -> Tuple[List, ...]:
        raise NotImplemented()

    def append_args(self, *row) -> 'Buffer':
        return self.append_row(row)

    def append_row(self, row: np.ndarray) -> 'Buffer':
        pass

    def reset(self) -> 'Buffer':
        pass

    def __getitem__(self, item):
        # return the column requested
        return None

    def __len__(self):
        return 0


class ListBuffer(Buffer):

    def __init__(self, columns: Union[int, Iterable[str]] = 1, max_size:int = None):
        self.column_names = {col: i for i, col in enumerate(list(range(columns)) if isinstance(columns, int) else columns)}
        self.columns = [[] for _ in range(len(self.column_names))]
        self.max_size = None
        self.push_count = 0

    def append_row(self, row: np.ndarray) -> 'ListBuffer':
        assert len(row) == len(self.columns)
        for i, col in enumerate(row):
            if self.max_size is not None and len(self) > self.max_size:
                self.columns[self.push_count % self.max_size].append(col)
            else:
                self.columns[i].append(col)

        self.push_count += 1
        return self

    def __getitem__(self, item):
        return self.columns[item if isinstance(item, int) else self.column_names[item]]

    def sample(self, size) -> Tuple[List, ...]:
        indices = np.random.choice(len(self), size)
        return tuple([[c[i] for i in indices] for c in self.columns])

    def reset(self) -> 'ListBuffer':
        return ListBuffer(self.column_names.keys())

    def __len__(self):
        return len(self.columns[0])
